/***********************************************************
 * Zuse Institute Berlin (ZIB), Takustr. 7, 14195 Berlin
 *
 * Exercise for implementing a dynamic program to solve the
 * constrained-shortest-path problem.
 * programming hints:
 *
 *  - compile and link with g++ rcsp.cpp -o rcsp
 *  - run with ./rcsp dutch.dat
 *
 ***********************************************************/

// c includes
#include <cstdio>
#include <cstdlib>
#include <cassert>
#include <cstring>
#include <climits>

// c++ includes
#include <fstream>
#include <iostream>
#include <sstream>
#include <string>
#include <map>


#define MAX(x,y)  ( (x) >= (y) ? (x) : (y) )
#define MIN(x,y)  ( (x) <= (y) ? (x) : (y) )

// define for infinity
#define INFTY  INT_MAX/2

typedef struct graph
{
  int n;             // number of nodes in the graph

  // reverse star graph implementation
  int *pred;         // stores all the predecessornodes
  int *count_pred;   // counts the predecessors for node i
  int *begin;        // stores where the predecessors for node i starts in pred[]

  int *time ;        // travel-time per arc
  int *cost ;        // cost per arc
  int *cap  ;        // capacity per arc (not used in this excercise)

} graph;

// map for mapping the names of the nodes with indices
std::map<std::string,unsigned int> vertices;

// map for mapping the indices of the nods with names
std::map<unsigned int,std::string> names;

void readgraph(char * filename, graph *g)
{
  using namespace std;

  unsigned int line = 0;

  unsigned int nodeCnt = 0 ;
  unsigned int arcCnt  = 0 ;

  ifstream fin;
  string buffer;

  fin.open(filename, ios_base::in);

  if( fin.fail() )
  {
    printf("Can't read file %s.\n", filename);
    fin.close();
    exit(1);
  }

  while (fin.good())
  {
    getline(fin, buffer, '\n');

    if(buffer.size() <= 1) continue;

    stringstream ss( buffer );

    string tmp;

    ss >> tmp ;

    assert(tmp.size() == 1);

    switch(tmp.at(0))
    {
      case 'N' :
      {
        int n;

        ss >> n ;

        g->pred       = (int*) calloc(n * n , sizeof(int)); // upper bound for the number of arcs
        g->count_pred = (int*) calloc(n , sizeof(int));     // allocate memorycapacity to
        g->begin      = (int*) calloc(n , sizeof(int));     // store the arcs of our graph

        g->time = (int*) calloc(n * n , sizeof(int));
        g->cost = (int*) calloc(n * n , sizeof(int));
        g->cap  = (int*) calloc(n * n , sizeof(int));

        for(int i = 0; i < n; i++ )
        {
          g->count_pred[i] = 0 ;
        }

        g->n = n ;

        for(int i = 0; i < n ; i++ )
          g->time[i] = g->cost[i] = g->cap[i] = 0 ;

        break;
      }
      case 'V' :
      {
        ss >> tmp;

        vertices[tmp]  = nodeCnt ;
        names[nodeCnt] = tmp ;

        nodeCnt++;

        break;
      }
      case 'A' :
      {
        string tailName, headName;

        int cap, time, cost;

        ss >> tailName >> headName >> cap >> time >> cost;

        int tail = vertices[tailName];
        int head = vertices[headName];

        if( g->count_pred[head] == 0 )
          g->begin[head] = arcCnt;

        g->count_pred[head]++ ;

        g->pred[arcCnt] = tail ;

        g->cap[arcCnt]  = cap;
        g->time[arcCnt] = time;
        g->cost[arcCnt] = cost;

        arcCnt++;

        break ;
      }
      case '#' :
      {
        break;
      }
    }

    line++;
  }

  fin.close();
}

void printGraph(graph * g)
{
  printf("nodes\n");

  for(int i = 0; i < g->n; i++ )
  {
    printf("node %2d (%s)\n",i,names[i].c_str());
  }

  printf("edges\n");

  for(int i = 0; i < g->n; i++ )
  {
    for(int j = g->begin[i]; j < g->begin[i] + g->count_pred[i]; j++ )
    {
      int pred = g->pred[j];

      printf("arc %2d (%4s) --> %2d (%4s) %6d %2d %6d\n",
          pred, names[pred].c_str(), i,names[i].c_str(),
          g->cap[j],g->time[j],g->cost[j]);
    }
  }
}

void printTable(int ** table,int n, int d,int r, int v, int rlb, int rub)
{
  printf("r=%d v=%d\n",r,v);
  for(int i =0; i < n; i++)
  {
    for(int j =0; j < d; j++)
      if(rlb <= j && rub >= j)
      {

      if(table[i][j] == INFTY)
        printf("%5s ","M");
      else
        printf("%5d ",table[i][j]);
      }
    printf("\n");
  }

  printf("\n");
}

void cspDP(graph * g, int s, int t, int d)
{
  int ** table = (int**) calloc(g->n,sizeof(int*));

  for(int i = 0 ; i < g->n; i++)
    table[i] = (int*) calloc(d,sizeof(int));

  for(int r = 0 ; r <= d; r++)
  {
    table[s][r] = 0 ;
  }

  for( int v = 0; v < g->n; v++ )
  {
    if (v != s)
      table[v][0] = INFTY;
  }

  for( int r = 1; r <= d; r++ )
    for( int v = 0; v < g->n; v++ )
    {
      if (v != s)
      {
        int cmin = INT_MAX;

        for( int i = g->begin[v]; i < g->begin[v] + g->count_pred[v]; i++ )
        {
          assert(i < g->n * g->n);

          int u = g->pred[i];

          int c_uv = g->time[i];
          int w_uv = g->cost[i];

          assert(c_uv >= 0);
          assert(w_uv >= 0);

          if( w_uv > r ) continue;

          assert(table[v][r - w_uv] >= 0);
          assert(c_uv >= 0);

          if (table[u][r - w_uv] + c_uv < cmin)
          {
            cmin = table[u][r - w_uv] + c_uv;
            assert(cmin >= 0);
          }
        }

        assert(table[v][r-1] >= 0);
        assert(cmin >= 0);

        table[v][r] = MIN(table[v][r-1],cmin);
      }
    }

  free(table);

  printf("cost = %d\n",table[t][d]);
}


int main(int argc, char **argv)
{
  using namespace std;

  graph g;

  readgraph(argv[1],&g);
  //printGraph(&g);

  unsigned int s = vertices["Gn"];
  unsigned int t = vertices["Rtd"];

  cspDP(&g,s,t,20000);

  free(g.pred);
  free(g.count_pred);
  free(g.begin);
  free(g.cap);
  free(g.cost);
  free(g.time);

  return 0;
}

